@@ -1,28 +1,37 @@
 use pyo3::prelude::*;
 use pyo3::pyclass::CompareOp;
-use serde::{Deserialize, Serialize};
+use serde::ser::{SerializeSeq, Serializer};
+use serde::Serialize;
+use serde_json::{Map, Value};
+use sha2::{Digest, Sha256};
 use std::cmp::Ordering;
+use std::collections::{HashMap, HashSet};
+use std::fmt;
 use std::fs::{read_to_string, File};
-use std::hash::{Hash, Hasher};
 use std::io::prelude::*;
-use std::{
-    collections::{HashMap, HashSet},
-    fmt,
-};
 
 use crate::column_schema;
 use crate::column_schema::ColumnSchema;
 use crate::logical_types::LogicalTypes;
 use crate::primitive::Primitive;
+#[derive(Serialize)]
+struct FileAsset {
+    features: Vec<Feature>,
+    primitives: Vec<Primitive>,
+}
 
 #[pyclass]
-#[derive(Debug, Eq, Deserialize, Serialize, Clone)]
+#[derive(Debug, Eq, Serialize, Clone)]
 pub struct Feature {
     #[pyo3(get, set)]
     pub name: String,
+    pub id: String,
     pub schema: ColumnSchema,
     #[pyo3(get, set)]
+    #[serde(serialize_with = "child_serialize")]
     pub base_features: Vec<Feature>,
+    #[pyo3(get, set)]
+    #[serde(serialize_with = "serialize_primitive")]
     pub generating_primitive: Option<Primitive>,
 }
 
@@ -77,12 +86,94 @@ impl Feature {
         ))
     }
 
+    #[staticmethod]
+    fn save_features(features: Vec<Feature>, filename: &str) -> PyResult<()> {
+        Feature::write_many_to_file(&features, filename);
+
+        Ok(())
+    }
+
     fn __richcmp__(&self, other: &Feature, _op: CompareOp) -> PyResult<bool> {
         // println!("{:?} == {:?}", self, other);
         Ok(self.equals(other))
     }
 }
 
+fn rec(
+    features: &Vec<Feature>,
+    feature_set: &mut HashSet<Feature>,
+    primitive_set: &mut HashSet<Primitive>,
+) {
+    for f in features {
+        feature_set.insert(f.clone());
+        if let Some(p) = &f.generating_primitive {
+            primitive_set.insert(p.clone());
+        }
+        rec(&f.base_features, feature_set, primitive_set);
+    }
+}
+
+fn rec2(
+    f: Value,
+    feature_map: &HashMap<String, Map<String, Value>>,
+    primitive_map: &HashMap<String, Primitive>,
+) -> Feature {
+    let name = f
+        .as_object()
+        .unwrap()
+        .get("name")
+        .unwrap()
+        .as_str()
+        .unwrap()
+        .to_string();
+    let id = f
+        .as_object()
+        .unwrap()
+        .get("id")
+        .unwrap()
+        .as_str()
+        .unwrap()
+        .to_string();
+    let schema_raw = f.as_object().unwrap().get("schema").unwrap().clone();
+
+    let schema: ColumnSchema = serde_json::from_value(schema_raw).unwrap();
+    let base_features = f
+        .as_object()
+        .unwrap()
+        .get("base_features")
+        .unwrap()
+        .as_array()
+        .unwrap()
+        .clone();
+    let generating_primitive_raw = f.as_object().unwrap().get("generating_primitive").unwrap();
+
+    let generating_primitive = match generating_primitive_raw.clone() {
+        Value::String(id) => {
+            let primitive = primitive_map.get(&id.clone()).unwrap().clone();
+            Some(primitive)
+        }
+        Value::Null => None,
+        _ => panic!("Unexpected generating primitive"),
+    };
+
+    let base_features: Vec<Feature> = base_features
+        .iter()
+        .map(|x| {
+            let id = x.as_str().unwrap().to_string();
+            let feature = feature_map.get(&id).unwrap().clone();
+            rec2(Value::Object(feature), feature_map, primitive_map)
+        })
+        .collect();
+
+    Feature {
+        name,
+        id,
+        schema,
+        base_features,
+        generating_primitive,
+    }
+}
+
 impl Feature {
     pub fn new(
         name: String,
@@ -96,21 +187,71 @@ impl Feature {
             schema.expect(format!("Error unwrapping schema in {}", name).as_str())
         };
 
+        let base_features = base_features.unwrap_or(vec![]);
+
         Feature {
-            name,
+            name: name.clone(),
+            id: Feature::gen_id(&name, &base_features, &generating_primitive),
             schema: dtype,
-            base_features: base_features.unwrap_or(vec![]),
+            base_features,
             generating_primitive,
         }
     }
 
+    fn gen_id(
+        name: &String,
+        base_features: &Vec<Feature>,
+        generating_primitive: &Option<Primitive>,
+    ) -> String {
+        let mut hasher = Sha256::new();
+        if let Some(ref o) = generating_primitive {
+            hasher.update(o.id.clone());
+
+            let bf_ids = base_features
+                .iter()
+                .map(|x| x.id.clone())
+                .collect::<Vec<_>>();
+            if o.commutative {
+                let mut bf_ids = bf_ids;
+                bf_ids.sort();
+                for id in bf_ids {
+                    hasher.update(id);
+                }
+            } else {
+                for id in bf_ids {
+                    hasher.update(id);
+                }
+            }
+        } else {
+            hasher.update(name);
+        }
+
+        let r = format!("{:x}", hasher.finalize());
+
+        r
+    }
+
     pub fn equals(&self, other: &Feature) -> bool {
         // --> do I need this?
         self == other
     }
 
-    pub fn write_many_to_file(features: &Vec<Feature>, filename: String) {
-        let a = serde_json::to_string(&features);
+    pub fn flatten(features: &Vec<Feature>) -> (HashSet<Feature>, HashSet<Primitive>) {
+        let mut feature_set: HashSet<Feature> = HashSet::new();
+        let mut primitive_set: HashSet<Primitive> = HashSet::new();
+
+        rec(features, &mut feature_set, &mut primitive_set);
+        return (feature_set, primitive_set);
+    }
+
+    pub fn write_many_to_file(features: &Vec<Feature>, filename: &str) {
+        let (all_features, all_primitives) = Feature::flatten(features);
+
+        let fa = FileAsset {
+            features: all_features.into_iter().collect(),
+            primitives: all_primitives.into_iter().collect(),
+        };
+        let a = serde_json::to_string(&fa);
 
         match a {
             Ok(s) => {
@@ -139,10 +280,71 @@ impl Feature {
         }
     }
 
-    pub fn read_from_file(filename: String) -> Feature {
+    pub fn read_from_file(filename: &str) -> Vec<Feature> {
         let contents = read_to_string(filename).expect("Something went wrong reading the file");
-        let feature: Feature = serde_json::from_str(&contents[..]).unwrap();
-        return feature;
+        let asset: Value = serde_json::from_str(&contents[..]).unwrap();
+
+        let primitives_raw = asset
+            .as_object()
+            .unwrap()
+            .get("primitives")
+            .unwrap()
+            .clone();
+
+        let primitives: Vec<Primitive> = serde_json::from_value(primitives_raw).unwrap();
+
+        let primitive_map: HashMap<String, Primitive> = primitives
+            .iter()
+            .map(|x| (x.id.clone(), x.clone()))
+            .collect();
+
+        // dbg!(primitive_map);
+
+        let features = asset
+            .as_object()
+            .unwrap()
+            .get("features")
+            .unwrap()
+            .as_array()
+            .unwrap();
+
+        let feature_map: HashMap<String, Map<String, Value>> = features
+            .iter()
+            .map(|x| {
+                let obj = x.as_object().unwrap();
+                let id = obj.get("id").unwrap().as_str().unwrap().to_string();
+
+                (id, obj.clone())
+            })
+            .collect();
+
+        let hydrated_features: Vec<Feature> = features
+            .iter()
+            .map(|f| rec2(f.clone(), &feature_map, &primitive_map))
+            .collect();
+
+        hydrated_features
+    }
+}
+
+fn child_serialize<S>(x: &Vec<Feature>, s: S) -> Result<S::Ok, S::Error>
+where
+    S: Serializer,
+{
+    let mut seq = s.serialize_seq(Some(x.len()))?;
+    for element in x {
+        seq.serialize_element(&element.id)?;
+    }
+    seq.end()
+}
+
+fn serialize_primitive<S>(x: &Option<Primitive>, s: S) -> Result<S::Ok, S::Error>
+where
+    S: Serializer,
+{
+    match *x {
+        Some(ref value) => s.serialize_some(&value.id),
+        None => s.serialize_none(),
     }
 }
 
@@ -152,24 +354,6 @@ impl fmt::Display for Feature {
     }
 }
 
-impl Hash for Feature {
-    fn hash<H: Hasher>(&self, state: &mut H) {
-        match self.generating_primitive {
-            Some(ref p1) => {
-                p1.hash(state);
-
-                let mut v = self.base_features.clone();
-                v.sort();
-
-                v.hash(state);
-            }
-            None => {
-                self.name.hash(state);
-            }
-        }
-    }
-}
-
 impl PartialOrd for Feature {
     fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
         Some(self.cmp(other))
@@ -178,52 +362,13 @@ impl PartialOrd for Feature {
 
 impl Ord for Feature {
     fn cmp(&self, other: &Self) -> Ordering {
-        // TODO: is match the right construct here? primititves aren't used
-        match (
-            self.generating_primitive.clone(),
-            other.generating_primitive.clone(),
-        ) {
-            (Some(ref _p1), Some(ref _p2)) => self.base_features.cmp(&other.base_features),
-            _ => return self.name.cmp(&other.name),
-        }
+        self.id.cmp(&other.id)
     }
 }
 
 impl PartialEq for Feature {
     fn eq(&self, other: &Self) -> bool {
-        match (
-            self.generating_primitive.clone(),
-            other.generating_primitive.clone(),
-        ) {
-            (None, None) => self.name == other.name,
-            (Some(ref p1), Some(ref p2)) => {
-                if p1 != p2 {
-                    return false;
-                } else {
-                    let bf = &self.base_features;
-                    let other_bf = &other.base_features;
-
-                    let is_commutative = self
-                        .generating_primitive
-                        .as_ref()
-                        .unwrap()
-                        .clone()
-                        .commutative;
-
-                    let bf_equal = match is_commutative {
-                        true => {
-                            let s1: HashSet<Feature> = HashSet::from_iter(bf.iter().cloned());
-                            let s2: HashSet<Feature> = HashSet::from_iter(other_bf.iter().cloned());
-                            s1 == s2
-                        }
-                        false => bf == other_bf,
-                    };
-
-                    return bf_equal;
-                }
-            }
-            _ => return false,
-        }
+        self.id == other.id
     }
 }
 
@@ -275,71 +420,40 @@ pub fn generate_fake_features(n_features: i32) -> Vec<Feature> {
 
     for i in 0..(n_features - 1) {
         let name = format!("F_{}", i);
-        // let logical_type = LogicalTypes::Integer;
-        // let semantic_type = LogicalTypes::Numeric;
-        // let base_features = None;
-        // let feature = Feature::new(name, logical_type, semantic_type, base_features, None);
         features.push(create_feature!(name, "integer", "numeric"));
     }
 
     return features;
 }
 
-#[pyfunction]
-pub fn compare_featuresets(
-    features1: Vec<Feature>,
-    features2: Vec<Feature>,
-) -> (Vec<Feature>, Vec<Feature>) {
-    let features1: HashSet<Feature> = HashSet::from_iter(features1.iter().cloned());
-
-    let features2: HashSet<Feature> = HashSet::from_iter(features2.iter().cloned());
-
-    let diff1: Vec<Feature> = features1.difference(&features2).cloned().collect();
-
-    let diff2: Vec<Feature> = features2.difference(&features1).cloned().collect();
-
-    return (diff1, diff2);
-}
-
-pub fn print_set(features: &Vec<Feature>) {
-    println!("===");
-    match features.len() {
-        0 => println!("No difference"),
-        _ => {
-            println!("Features from s1 not in s2:");
-            features.iter().for_each(|f| {
-                println!("\t-> {}", f);
-            });
-        }
-    }
-
-    println!("===");
-}
-
 #[cfg(test)]
 mod tests {
-    use std::collections::{HashMap, HashSet};
+    use serde_json::{Map, Value};
 
-    use crate::{
-        column_schema,
-        feature::{compare_featuresets, hash_features},
-        input_set,
-        logical_types::LogicalTypes,
-        primitive,
+    use super::*;
+    use std::{
+        collections::{HashMap, HashSet},
+        fs,
     };
 
+    use crate::{column_schema, input_set, primitive};
+
     #[test]
     fn test_eq1() {
         let bf1 = create_feature!("bf1", "integer", "index");
         let bf1a = create_feature!("bf1", "integer", "index");
         let bf2 = create_feature!("bf2", "integer", "index");
 
+        // assert origin features with same name and datatypes are equal
         assert_eq!(bf1, bf1a);
+
+        // assert origin features with different names are not equal
         assert_ne!(bf1, bf2);
 
         assert_eq!([bf1.clone(), bf2.clone()], [bf1.clone(), bf2.clone()]);
         assert_ne!([bf1.clone(), bf2.clone()], [bf2.clone(), bf1.clone()]);
 
+        // assert we can hash features, and that equal features are condensed
         let s1 = HashSet::from([&bf1, &bf1a, &bf2]);
 
         println!("{:?}", s1.contains(&bf1));
@@ -349,8 +463,9 @@ mod tests {
 
     #[test]
     fn test_eq2() {
-        let bf1 = create_feature!("bf1", "integer", "index");
-        let bf2 = create_feature!("bf2", "integer", "index");
+        // Create base features
+        let bf1 = create_feature!("origin_1", "integer", "index");
+        let bf2 = create_feature!("origin_2", "integer", "index");
 
         let inputs1 = vec![
             input_set![["", "numeric"], ["", "numeric"]],
@@ -373,11 +488,6 @@ mod tests {
 
         let s1 = HashSet::from([&f1, &f2, &f3]);
 
-        // println!(
-        //     "{:?}",
-        //     s1.into_iter().map(|f| &f.name[..]).collect::<Vec<_>>()
-        // );
-
         assert!(s1.len() == 2);
 
         // ================================================================
@@ -400,10 +510,6 @@ mod tests {
 
         let s2 = HashSet::from([&f1a, &f2a, &f3a]);
 
-        // println!(
-        //     "{:?}",
-        //     s2.into_iter().map(|f| &f.name[..]).collect::<Vec<_>>()
-        // );
         assert!(s2.len() == 1);
 
         let c1 = column_schema!["double", "numeric"];
@@ -416,37 +522,6 @@ mod tests {
         assert!(s3.len() == 2);
     }
 
-    #[test]
-    fn test_compare_featuresets() {
-        let bf1 = create_feature!("bf1", "integer", "index");
-        let bf2 = create_feature!("bf2", "integer", "index");
-
-        let inputs1 = vec![
-            input_set![["", "numeric"], ["", "numeric"]],
-            input_set![["datetime", ""], ["datetime", ""]],
-            input_set![["ordinal", ""], ["ordinal", ""]],
-        ];
-
-        let r_type1 = column_schema!["integer", "numeric"];
-
-        // Commutative is false, so order matters
-        let p1 = primitive!("GreaterThan", inputs1, r_type1, false);
-
-        let f1 = create_feature!("f1", p1.clone(), [bf1.clone(), bf2.clone()]);
-        let f2 = create_feature!("f2", p1.clone(), [bf1.clone(), bf2.clone()]);
-        let f3 = create_feature!("f2", p1.clone(), [bf2.clone(), bf1.clone()]);
-
-        let (d1, d2) = compare_featuresets(vec![f1.clone()], vec![f2.clone()]);
-
-        assert!(d1.len() == 0);
-        assert!(d2.len() == 0);
-
-        let (d1, d2) = compare_featuresets(vec![f1.clone()], vec![f3.clone()]);
-
-        assert!(d1.len() == 1);
-        assert!(d2.len() == 1);
-    }
-
     #[test]
     fn test_macro() {
         let inputs = vec![
@@ -466,4 +541,119 @@ mod tests {
 
         println!("{:?}", f1);
     }
+
+    #[test]
+    fn test_serialize() {
+        let f = create_feature!("bf1", "integer", "index");
+
+        let json = serde_json::to_string(&f).unwrap();
+
+        let expected = r#"{"name":"bf1","id":"def64d725ecd4bc3fe5315e7eb56ac740603d1372bec6b8925527ba70b793c69","schema":{"logical_type":"Integer","semantic_tag":"Index"},"base_features":[],"generating_primitive":null}"#;
+
+        assert!(json == expected.to_string());
+
+        let bf1 = create_feature!("origin_1", "integer", "index");
+        let bf2 = create_feature!("origin_2", "integer", "index");
+
+        let inputs1 = vec![
+            input_set![["", "numeric"], ["", "numeric"]],
+            input_set![["datetime", ""], ["datetime", ""]],
+            input_set![["ordinal", ""], ["ordinal", ""]],
+        ];
+
+        let r_type1 = column_schema!["integer", "numeric"];
+
+        // Commutative is false, so order matters
+        // result should be ["F1", "F3"], since "F2" is same as "F1"
+        let p1 = primitive!("GreaterThan", inputs1, r_type1, false);
+
+        let f1 = create_feature!("f1", p1.clone(), [bf1.clone(), bf2.clone()]);
+
+        let json = serde_json::to_string(&f1).unwrap();
+
+        let expected = r#"{"name":"f1","id":"fa15e4309253d1cb1c5a9a9a52ff307fcdfc11402bd913d3f27b9d4b40424888","schema":{"logical_type":"Integer","semantic_tag":"Numeric"},"base_features":["760538b502bd28c2875ff2d17fd5e24749df190ff6e49ccc44267ea07f404641","c70faa957d52eb965690b7e1a9a4ef5b0ea02848322b9b4421f2924402b9610b"],"generating_primitive":"e3a64445d604e1b5aab469e2c4d9b7a9504e1b9960341d4357f19ba8fa89189d"}"#;
+
+        assert!(json == expected.to_string());
+    }
+
+    #[test]
+    fn test_flatten() {
+        let bf1 = create_feature!("origin_1", "integer", "index");
+        let bf2 = create_feature!("origin_2", "integer", "index");
+
+        let inputs1 = vec![
+            input_set![["", "numeric"], ["", "numeric"]],
+            input_set![["datetime", ""], ["datetime", ""]],
+            input_set![["ordinal", ""], ["ordinal", ""]],
+        ];
+
+        let r_type1 = column_schema!["integer", "numeric"];
+
+        // Commutative is false, so order matters
+        // result should be ["F1", "F3"], since "F2" is same as "F1"
+        let p1 = primitive!("GreaterThan", inputs1, r_type1, false);
+
+        let f1 = create_feature!("f1", p1.clone(), [bf1.clone(), bf2.clone()]);
+
+        let features = vec![f1];
+        let (all_features, all_primitives) = Feature::flatten(&features);
+
+        assert_eq!(all_features.len(), 3);
+        assert_eq!(all_primitives.len(), 1);
+    }
+
+    #[test]
+    fn test_save() {
+        let test_filename = "test_features.json";
+        let bf1 = create_feature!("origin_1", "integer", "index");
+        let bf2 = create_feature!("origin_2", "integer", "index");
+
+        let inputs1 = vec![
+            input_set![["", "numeric"], ["", "numeric"]],
+            input_set![["datetime", ""], ["datetime", ""]],
+            input_set![["ordinal", ""], ["ordinal", ""]],
+        ];
+
+        let r_type1 = column_schema!["integer", "numeric"];
+
+        // Commutative is false, so order matters
+        // result should be ["F1", "F3"], since "F2" is same as "F1"
+        let p1 = primitive!("GreaterThan", inputs1, r_type1, false);
+
+        let f1 = create_feature!("f1", p1.clone(), [bf1.clone(), bf2.clone()]);
+
+        let features = vec![f1];
+        Feature::write_many_to_file(&features, test_filename);
+        fs::remove_file(test_filename).unwrap();
+    }
+
+    #[test]
+    fn test_read() {
+        let test_filename = "test_features2.json";
+        let bf1 = create_feature!("origin_1", "integer", "index");
+        let bf2 = create_feature!("origin_2", "integer", "index");
+
+        let inputs1 = vec![
+            input_set![["", "numeric"], ["", "numeric"]],
+            input_set![["datetime", ""], ["datetime", ""]],
+            input_set![["ordinal", ""], ["ordinal", ""]],
+        ];
+
+        let r_type1 = column_schema!["integer", "numeric"];
+
+        // Commutative is false, so order matters
+        // result should be ["F1", "F3"], since "F2" is same as "F1"
+        let p1 = primitive!("GreaterThan", inputs1, r_type1, false);
+
+        let f1 = create_feature!("f1", p1.clone(), [bf1.clone(), bf2.clone()]);
+
+        let features = vec![f1];
+        Feature::write_many_to_file(&features, test_filename);
+
+        let features = Feature::read_from_file(test_filename);
+
+        fs::remove_file(test_filename).unwrap();
+
+        assert_eq!(features.len(), 3);
+    }
 }
