@@ -1,5 +1,8 @@
 mod feature;
+mod hash;
 mod hash_features;
+mod utils;
 
 pub use self::feature::*;
 pub use self::hash_features::*;
+pub use self::utils::*;
